@@ -650,8 +650,8 @@ int     smtp_helo(SMTP_STATE *state)
      * non-SMTPUTF8 server? That could make life easier for mailing lists.
      */
 #define DELIVERY_REQUIRES_SMTPUTF8 \
-	((request->smtputf8 & SMTPUTF8_FLAG_REQUESTED) \
-	&& (request->smtputf8 & ~SMTPUTF8_FLAG_REQUESTED))
+	((request->sendopts & SMTPUTF8_FLAG_REQUESTED) \
+	&& (request->sendopts & SMTPUTF8_FLAG_DERIVED))
 
     /*
      * Require that the server supports SMTPUTF8 when delivery requires
@@ -926,13 +926,16 @@ static int smtp_start_tls(SMTP_STATE *state)
      * XXX: The TLS library will salt the serverid with further details of the
      * protocol and cipher requirements including the server ehlo response.
      * Deferring the helo to the digested suffix results in more predictable
-     * SSL session lookup key lengths.
+     * SSL session lookup key lengths. Add the current TLS security level to
+     * account for TLS level overrides based on message content or envelope
+     * metadata.
      */
     serverid = vstring_alloc(10);
     smtp_key_prefix(serverid, "&", state->iterator, SMTP_KEY_FLAG_SERVICE
 		    | SMTP_KEY_FLAG_CUR_NEXTHOP	/* With port */
 		    | SMTP_KEY_FLAG_HOSTNAME
-		    | SMTP_KEY_FLAG_ADDR);
+		    | SMTP_KEY_FLAG_ADDR
+		    | SMTP_KEY_FLAG_TLS_LEVEL);
 
     if (state->tls->conn_reuse) {
 	TLS_CLIENT_PARAMS tls_params;
@@ -1776,8 +1779,9 @@ static int smtp_loop(SMTP_STATE *state, NOCLOBBER int send_state,
 	     * the SMTPUTF8 promise that was made to the sender.
 	     */
 	    if ((session->features & SMTP_FEATURE_SMTPUTF8) != 0
-		&& (request->smtputf8 & SMTPUTF8_FLAG_REQUESTED) != 0)
+		&& (request->sendopts & SMTPUTF8_FLAG_REQUESTED) != 0)
 		vstring_strcat(next_command, " SMTPUTF8");
+	    /* TODO(wietse) REQUIRETLS. */
 
 	    /*
 	     * We authenticate the local MTA only, but not the sender.
@@ -1841,8 +1845,9 @@ static int smtp_loop(SMTP_STATE *state, NOCLOBBER int send_state,
 		    quote_822_local(session->scratch, rcpt->orig_addr);
 		    vstring_sprintf(session->scratch2, "%s;%s",
 		    /* Fix 20140707: sender must request SMTPUTF8. */
-				    (request->smtputf8 != 0
-			      && !allascii(vstring_str(session->scratch))) ?
+				    ((request->sendopts & SMTPUTF8_FLAG_ALL)
+				 && !allascii(vstring_str(session->scratch))
+		     && valid_utf8_stringz(vstring_str(session->scratch))) ?
 				    "utf-8" : "rfc822",
 				    vstring_str(session->scratch));
 		    orcpt_type_addr = vstring_str(session->scratch2);
